from abc import ABCMeta
from dataclasses import dataclass
from functools import cached_property
from typing import Optional

from ..installed_files import Tracker
from ..state import Storage, storage
from .base import Package


@dataclass
class Version:
    version: str


VERSIONS = storage("version", Version)


class ManualVersion(Package, metaclass=ABCMeta):
    @cached_property
    def versions(self) -> Storage[Version]:
        return VERSIONS(self.db)

    @property
    def cached_version(self) -> Optional[str]:
        try:
            return self.versions[self.name].version
        except KeyError:
            return None

    async def local_version(self) -> Optional[str]:
        return self.cached_version

    async def cache_version(self) -> None:
        self.versions[self.name] = Version(version=await self.get_remote_version())

    async def install(self, *, tracker: Tracker) -> None:
        await super().install(tracker=tracker)
        await self.cache_version()



from contextlib import asynccontextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import AsyncIterator

from .driver import Driver
from .state import DB, Storage, storage


@dataclass(frozen=True)
class InstalledFile:
    path: str
    root: bool

    @property
    def path_(self) -> Path:
        return Path(self.path)


INSTALLED_FILES = storage("installed_file", InstalledFile)


@dataclass
class Tracker:
    storage: Storage[InstalledFile]
    current: set[InstalledFile]
    previous: set[InstalledFile]

    def track(self, target: Path, *, root: bool = False) -> None:
        installed = InstalledFile(path=str(target), root=root)
        if installed not in self.current:
            self.current.add(installed)
            if installed not in self.previous:
                self.storage.append(installed)


@asynccontextmanager
async def track_files(db: DB, driver: Driver) -> AsyncIterator[Tracker]:
    installed_files = INSTALLED_FILES(db)

    current: set[InstalledFile] = set()
    previous: set[InstalledFile] = set(installed_files.find())

    yield Tracker(
        storage=installed_files,
        current=current,
        previous=previous,
    )

    for removed in previous - current:
        await driver.with_root(removed.root).rm(removed.path_)
        installed_files.delete(path=removed.path)
